from pathlib import Path
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import time
import sys
import shutil

from pyfiglet import Figlet
import transformers
from halo import Halo

from .args import setup_cmdline_args_for_docly_gen
from .setup_env import inspect_and_download_latest_model, inspect_and_download_latest_tslibs
from docly.config import DoclyConfig
from docly.ioutils import (is_dir,
                           check_out_path,
                           process_file,
                           is_python_file,
                           is_ipynb_notebook,
                           query_yes_no,
                           look_for_update
                           )
from docly.ioutils.console_printer import print_on_console
from docly.ioutils.apply_diff import apply_diff
from docly.ioutils.table_printer import print_results_as_table
from docly.ioutils.convert_ipynb import convert_ipynb_to_python
from docly.logic.logic_main import load_model, predict_docstring

transformers.logger.setLevel(transformers.logging.CRITICAL)

parser = ArgumentParser()
ROOT = (Path.home() / ".docly")
MODEL_DOWNLOAD_ROOT = "https://docly-model.s3.amazonaws.com/pytorch_model.bin"
NEW_MODEL_DOWNLOAD_ROOT = "https://code-summary.s3.amazonaws.com/pytorch_model_new.bin"
TSLIBS_DOWNLOAD_ROOT = "https://func2docstr-py.s3.amazonaws.com/"

f = Figlet(font='slant')


def _print_welcome():
    print(f.renderText('Docly'))


# def _predict_docstrings_from_file(f_path: Path, ts_lib_path: str, model: object, tokenizer: object):
#     func_names = []
#     ct = []
#     for code_tokens, raw_code, start_index, function_name in process_file(f_path, ts_lib_path):
#         func_names.append(func_names)
#         ct.append(code_tokens)
#     docstrs = predict_docstring(model, tokenizer, ct)
#     print(docstrs)


def _apply_diff(docstr_loc, no_generate_args_list, ipynb_files, docstring_style):
    print_on_console("Applying diff", color="green")
    apply_diff(docstr_loc, no_generate_args_list, ipynb_files, docstring_style)
    print_on_console("Diff applied. Good bye!", color="green", emoji="thumbsup")


def _remove_converted_python_files(ipynb_files):
    for py_file_loc, _ in ipynb_files.items():
        Path(py_file_loc).unlink()


def _if_jupytext_is_installed():
    try:
        import jupytext
        return True
    except ModuleNotFoundError:
        return False


def _deal_with_result(args, table_rows, docstr_loc, ipynb_files):
    if not args.no_generate_diff and table_rows:
        print_results_as_table(table_rows)
    elif args.force and docstr_loc:
        _apply_diff(docstr_loc, args.no_generate_args_list, ipynb_files, args.docstring_style)
        _remove_converted_python_files(ipynb_files)
    elif args.no_generate_diff and docstr_loc:
        if args.no_print_report:
            choice = query_yes_no("The diff has been generated, do you want to see the suggestions for missing Docstrings?")
            if choice:
                print_results_as_table(table_rows)
                choice = query_yes_no("Do you want to apply the suggestions?")
            else:
                choice = query_yes_no("Do you want to apply the suggestions?")
        else:
            choice = query_yes_no("Do you want to apply the suggestions?")
        
        if choice:
            _apply_diff(docstr_loc, args.no_generate_args_list, ipynb_files, args.docstring_style)
        else:
            _remove_converted_python_files(ipynb_files)
            print_on_console("Nothing changed. Good bye!", color="green", emoji="thumbsup")
    else:
        print_on_console("\n\nNothing to be done. Good bye!", color="green", emoji="thumbsup")


@Halo(text='Processing files', spinner='dots')
def _process(args, model, tokenizer, ts_lib_path, config: DoclyConfig):
    """
    Terribly written code. Refactor ASAP
    """
    table_rows = []
    docstr_loc = {}  # Very badly named variable. Need to change
    ipynb_files = {}

    for file in args.files:
        f_path = Path(file)
        if is_dir(file):
            for f in check_out_path(f_path):
                if not is_dir(f) and is_python_file(f):
                    ####
                    # Very bad implementation. Change ASAP
                    ####
                    if not config.is_dir_skipped(str(f).split("/")[:-1]):
                        for code_tokens, params, start_index, function_name, ds in process_file(f, ts_lib_path, args.use_old_model):
                            if ds == "":
                                docstr = predict_docstring(model, tokenizer, code_tokens, args.use_old_model)
                                if docstr_loc.get(str(f)) is None:
                                    docstr_loc[str(f)] = {start_index[0]: 
                                                                (start_index[1], 
                                                                docstr[0], 
                                                                params
                                                                )
                                                         }
                                else:
                                    docstr_loc[str(f)][start_index[0]] = (start_index[1], 
                                                                          docstr[0],
                                                                          params)
                                table_rows.append([f.name, function_name, docstr[0]])
                elif not is_dir(f) and is_ipynb_notebook(f) and args.run_on_notebooks:
                    if not config.is_dir_skipped(str(f).split("/")[:-1]):
                        py_file = convert_ipynb_to_python(f)
                        if py_file:
                            for code_tokens, params, start_index, function_name, ds in process_file(py_file, ts_lib_path, args.use_old_model):
                                if ds == "":
                                    docstr = predict_docstring(model, tokenizer, code_tokens, args.use_old_model)
                                    if docstr_loc.get(str(py_file)) is None:
                                            docstr_loc[str(py_file)] = {start_index[0]: 
                                                                            (start_index[1], 
                                                                            docstr[0], 
                                                                            params
                                                                            )
                                                                        }
                                    else:
                                        docstr_loc[str(py_file)][start_index[0]] = (start_index[1], 
                                                                                    docstr[0],
                                                                                    params)
                                    table_rows.append([f.name, function_name, docstr[0]])
                                    ipynb_files[str(py_file.absolute())] = f
        else:
            if is_python_file(f_path):
                if not config.is_dir_skipped(str(f_path.absolute()).split("/")[:-1]):
                    for code_tokens, params, start_index, function_name, ds in process_file(f_path, ts_lib_path, args.use_old_model):
                        if ds == "":
                            docstr = predict_docstring(model, tokenizer, code_tokens, args.use_old_model)
                            if docstr_loc.get(str(f_path.absolute())) is None:
                                docstr_loc[str(f_path.absolute())] = {start_index[0]: 
                                                                        (start_index[1], 
                                                                        docstr[0],
                                                                        params
                                                                        )
                                                                     }
                            else:
                                docstr_loc[str(f_path.absolute())][start_index[0]] = (start_index[1], 
                                                                                      docstr[0],
                                                                                      params)
                            table_rows.append([f_path.name, function_name, docstr[0]])
            elif is_ipynb_notebook(f_path) and args.run_on_notebooks:
                if not config.is_dir_skipped(str(f_path.absolute()).split("/")[:-1]):
                    py_file = convert_ipynb_to_python(f_path)
                    if py_file:
                        for code_tokens, params, start_index, function_name, ds in process_file(py_file, ts_lib_path, args.use_old_model):
                            if ds == "":
                                docstr = predict_docstring(model, tokenizer, code_tokens, args.use_old_model)
                                if docstr_loc.get(str(py_file)) is None:
                                        docstr_loc[str(py_file)] = {start_index[0]: 
                                                                        (start_index[1], 
                                                                        docstr[0],
                                                                        params
                                                                        )
                                                                    }
                                else:
                                    docstr_loc[str(py_file)][start_index[0]] = (start_index[1], 
                                                                                docstr[0],
                                                                                params)
                                table_rows.append([f_path.name, function_name, docstr[0]])
                                ipynb_files[str(py_file.absolute())] = f_path.absolute()
    return table_rows, docstr_loc, ipynb_files



def main():
    if look_for_update():
        print_on_console("There is an update available. Please run `pip install --upgrade docly`", color="green", emoji="rotating_light")
    _print_welcome()
    
    setup_cmdline_args_for_docly_gen(parser)
    args = parser.parse_args()

    if args.run_on_notebooks and not _if_jupytext_is_installed():
        print_on_console("You have mentioned `run_on_notebooks` but the needed dependecy is not present. Please run `pip install 'docly[jupyter]'` for that. This switch will be ignored", color="green")
        args.run_on_notebooks = False
    
    # if args.run_on_notebooks:
    #     print_on_console("You have mentioned the `run_on_notebooks` switch. It is experimental", color="red", emoji="rotating_light")
    #     choice = query_yes_no("Do you want to continue?")
    #     if not choice:
    #         args.run_on_notebooks = False

    config = DoclyConfig(args.config_file)

    try:
        mdp = NEW_MODEL_DOWNLOAD_ROOT if not args.use_old_model else MODEL_DOWNLOAD_ROOT
        inspect_and_download_latest_model(ROOT, mdp, args.use_old_model)
    except KeyboardInterrupt:
        print_on_console("You stopped the download. Docly won't work", color="red", emoji="X")
        shutil.rmtree(str(ROOT / "model"))
        sys.exit(1)
    
    try:
        ready, tslib_file = inspect_and_download_latest_tslibs(ROOT, TSLIBS_DOWNLOAD_ROOT)
        if not ready:
            print_on_console("===== OS version not supported =====", color="red", emoji="X")
            return
    except KeyboardInterrupt:
        print_on_console("You stopped the download. Docly won't work", color="red", emoji="X")
        sys.exit(1)

    print_on_console("Loading Engine. Please wait", color="green")
    model_name = "pytorch_model_new.bin" if not args.use_old_model else "pytorch_model.bin"
    model, tokenizer = load_model(str(ROOT / "model"/ model_name), args.use_old_model)
    print_on_console("Engine Loaded.", color="green", emoji="heavy_check_mark")
    
    ts_lib_path = str(ROOT / "tslibs" / tslib_file)
    
    table_rows, docstr_loc, ipynb_files = _process(args, model, tokenizer, ts_lib_path, config)

    _deal_with_result(args, table_rows, docstr_loc, ipynb_files)



from rich.table import Table
from rich.console import Console


def print_results_as_table(rows):
    table = Table(title="Functions and Docstrings")

    table.add_column("File Name", justify="left")
    table.add_column("Function Name", justify="left")
    table.add_column("Docstring", justify="left")

    for row in rows:
        table.add_row(*row)
    
    console = Console()
    console.print(table)



#########################

# This entire file should be re-written 
# using ast or tree-hugger tree traversal
# The way it is now, can cause problems 
# in some corner cases.

#######################

import os
import shutil
from pathlib import Path
from typing import Dict
# from tabnanny import check

from docly.ioutils import CACHE_DIR, cache_exists, make_cache_dir
from docly.ioutils.convert_ipynb import convert_python_to_ipynb


def _generate_main_docstring(docstring, spaces, style, will_follow_param: bool):
    if not will_follow_param:
        if style == "google":
            return f'{spaces}"""\n{spaces}{docstring}\n\n{spaces}(Generated by docly)\n{spaces}"""\n'
        elif style == "sphinx":
            return f'{spaces}"""{docstring}\n\n{spaces}(Generated by docly)\n{spaces}"""\n'
        elif style == "numpy":
            return f"{spaces}'''\n{spaces}{docstring}\n\n{spaces}(Generated by docly)\n{spaces}'''\n"
    else:
        if style == "google":
            return f'{spaces}"""\n{spaces}{docstring}'
        elif style == "sphinx":
            return f'{spaces}"""{docstring}'
        elif style == "numpy":
            return f"{spaces}'''\n{spaces}{docstring}"


def _generate_param_list(params, style, spaces):
    line_to_write = ""
    if style == "google":
        line_to_write = line_to_write + f"\n\n{spaces}Args:\n"
        for (param_name, param_type, default_val) in params:
            param_desc = f"{param_name}"
            if param_type:
                param_desc = param_desc + f" ({param_type})"
            if default_val:
                param_desc = param_desc + f" : Defaults to {default_val}"
            if param_desc.find(":") == -1:
                param_desc = param_desc + " :"
            if len(spaces) == 4:
                line_to_write = line_to_write + f"{spaces}{spaces}{param_desc}\n"
            if len(spaces) > 4:
                line_to_write = line_to_write + f"{spaces}    {param_desc}\n"
    elif style == "numpy":
        line_to_write = line_to_write + f"\n\n{spaces}Parameters\n{spaces}----------"
        for (param_name, param_type, default_val) in params:
            param_desc = f"\n{spaces}{param_name} :"
            if param_type:
                param_desc = param_desc + f" {param_type}"
            line_to_write = line_to_write + f"{param_desc}"
    elif style == "sphinx":
        line_to_write = "\n"
        for (param_name, param_type, default_val) in params:
            param_desc = f"\n{spaces}:param {param_name}:"
            if default_val:
                param_desc = param_desc + f"\n{spaces}    defaults to {default_val}"
            if param_type:
                param_desc = param_desc + f"\n{spaces}:type {param_name}: {param_type}"
            line_to_write = line_to_write + f"{param_desc}"
    return line_to_write


def _get_line_to_write(docstrs, line_num, should_write_args_list: bool, docstring_style: str):
    docstr_line = docstrs.get(line_num+1)[1]
    num_spaces = int(docstrs.get(line_num+1)[0])
    params = docstrs.get(line_num+1)[2]
    
    line_to_write = None

    spaces = " ".join([''] * (num_spaces + 1))
    if not params:
        line_to_write = _generate_main_docstring(docstr_line, spaces, docstring_style, False)
    else:
        line_to_write = _generate_main_docstring(docstr_line, spaces, docstring_style, True)
        if should_write_args_list:
            line_to_write = line_to_write + _generate_param_list(params, docstring_style, spaces)
        if docstring_style == "google" or docstring_style == "sphinx":
            line_to_write = line_to_write + f'\n\n{spaces}(Generated by docly)\n{spaces}"""\n'
        elif docstring_style == "numpy":
            line_to_write = line_to_write + f"\n\n{spaces}(Generated by docly)\n{spaces}'''\n"
    return line_to_write


def apply_diff(docstr_loc: Dict[str, Dict[int, tuple]], should_write_args_list: bool, ipynb_files: Dict, docstring_style: str):
    try:
        for file_loc, docstrs in docstr_loc.items():
            # l = check(file_loc)
            temp_file_name = f"{str(Path(file_loc).stem)}.pytemp"
            final_file_name = f"{str(Path(file_loc).stem)}.py"
            temp_file = (Path(file_loc).parent / temp_file_name)
            final_file = (Path(file_loc).parent / final_file_name)
            write_handle = open(temp_file, "w")
            
            with open(file_loc) as f:
                for line_num, line in enumerate(f):
                    if docstrs.get(line_num+1):
                        line_to_write = _get_line_to_write(docstrs, line_num, should_write_args_list, docstring_style)
                        write_handle.write(line_to_write)
                        write_handle.write(line)
                    else:
                        write_handle.write(line)
            
            write_handle.close()
            if len(ipynb_files) == 0:
                cache_file_name = file_loc[1:].replace("/", "#")
                
                if not cache_exists():
                    make_cache_dir()
                
                if (CACHE_DIR / cache_file_name).exists():
                    (CACHE_DIR / cache_file_name).unlink()
                
                shutil.move(file_loc, str(CACHE_DIR / cache_file_name))
                shutil.move(str(temp_file), str(final_file))
            elif len(ipynb_files) > 0 and file_loc not in ipynb_files.keys():
                cache_file_name = file_loc[1:].replace("/", "#")
                
                if not cache_exists():
                    make_cache_dir()
                
                if (CACHE_DIR / cache_file_name).exists():
                    (CACHE_DIR / cache_file_name).unlink()
                
                shutil.move(file_loc, str(CACHE_DIR / cache_file_name))
                shutil.move(str(temp_file), str(final_file))
            else:
                if Path(final_file).exists():
                    Path(final_file).unlink()
                
                shutil.move(str(temp_file), str(final_file))
                convert_python_to_ipynb(Path(final_file))

                Path(final_file).unlink()
    except KeyboardInterrupt:
        temp_file.unlink()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import torch
import torch.nn as nn


class Seq2Seq(nn.Module):
    """
    Build Seqence-to-Sequence.

    Parameters:

    * `encoder`- encoder of seq2seq model. e.g. roberta
    * `decoder`- decoder of seq2seq model. e.g. transformer
    * `config`- configuration of encoder model.
    * `beam_size`- beam size for beam search.
    * `max_length`- max length of target for beam search.
    * `sos_id`- start of symbol ids in target for beam search.
    * `eos_id`- end of symbol ids in target for beam search.
    """

    def __init__(
        self,
        encoder,
        decoder,
        config,
        beam_size=None,
        max_length=None,
        sos_id=None,
        eos_id=None,
    ):
        super(Seq2Seq, self).__init__()
        self.encoder = encoder
        self.decoder = decoder
        self.config = config
        self.register_buffer("bias", torch.tril(torch.ones(2048, 2048)))
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)
        self.lsm = nn.LogSoftmax(dim=-1)
        self.tie_weights()

        self.beam_size = beam_size
        self.max_length = max_length
        self.sos_id = sos_id
        self.eos_id = eos_id

    def _tie_or_clone_weights(self, first_module, second_module):
        """Tie or clone module weights depending of weither we are using TorchScript or not"""
        if self.config.torchscript:
            first_module.weight = nn.Parameter(second_module.weight.clone())
        else:
            first_module.weight = second_module.weight

    def tie_weights(self):
        """Make sure we are sharing the input and output embeddings.
        Export to TorchScript can't handle parameter sharing so we are cloning them instead.
        """
        self._tie_or_clone_weights(
            self.lm_head, self.encoder.embeddings.word_embeddings
        )

    def forward(
        self,
        source_ids=None,
        source_mask=None,
        target_ids=None,
        target_mask=None,
        args=None,
    ):
        outputs = self.encoder(source_ids, attention_mask=source_mask)
        encoder_output = outputs[0].permute([1, 0, 2]).contiguous()
        if target_ids is not None:
            attn_mask = -1e4 * (
                1 - self.bias[: target_ids.shape[1], : target_ids.shape[1]]
            )
            tgt_embeddings = (
                self.encoder.embeddings(target_ids).permute([1, 0, 2]).contiguous()
            )
            out = self.decoder(
                tgt_embeddings,
                encoder_output,
                tgt_mask=attn_mask,
                memory_key_padding_mask=(1 - source_mask).bool(),
            )
            hidden_states = torch.tanh(self.dense(out)).permute([1, 0, 2]).contiguous()
            lm_logits = self.lm_head(hidden_states)
            # Shift so that tokens < n predict n
            active_loss = target_mask[..., 1:].ne(0).view(-1) == 1
            shift_logits = lm_logits[..., :-1, :].contiguous()
            shift_labels = target_ids[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = nn.CrossEntropyLoss(ignore_index=-1)
            loss = loss_fct(
                shift_logits.view(-1, shift_logits.size(-1))[active_loss],
                shift_labels.view(-1)[active_loss],
            )

            outputs = loss, loss * active_loss.sum(), active_loss.sum()
            return outputs
        else:
            # Predict
            preds = []
            if source_ids.device.type == "cuda":
                zero = torch.cuda.LongTensor(1).fill_(0)
            elif source_ids.device.type == "cpu":
                zero = torch.LongTensor(1).fill_(0)
            for i in range(source_ids.shape[0]):
                context = encoder_output[:, i : i + 1]
                context_mask = source_mask[i : i + 1, :]
                beam = Beam(
                    self.beam_size,
                    self.sos_id,
                    self.eos_id,
                    device=source_ids.device.type,
                )
                input_ids = beam.getCurrentState()
                context = context.repeat(1, self.beam_size, 1)
                context_mask = context_mask.repeat(self.beam_size, 1)
                for _ in range(self.max_length):
                    if beam.done():
                        break
                    attn_mask = -1e4 * (
                        1 - self.bias[: input_ids.shape[1], : input_ids.shape[1]]
                    )
                    tgt_embeddings = (
                        self.encoder.embeddings(input_ids)
                        .permute([1, 0, 2])
                        .contiguous()
                    )
                    out = self.decoder(
                        tgt_embeddings,
                        context,
                        tgt_mask=attn_mask,
                        memory_key_padding_mask=(1 - context_mask).bool(),
                    )
                    out = torch.tanh(self.dense(out))
                    hidden_states = out.permute([1, 0, 2]).contiguous()[:, -1, :]
                    out = self.lsm(self.lm_head(hidden_states)).data
                    beam.advance(out)
                    input_ids.data.copy_(
                        input_ids.data.index_select(0, beam.getCurrentOrigin())
                    )
                    input_ids = torch.cat((input_ids, beam.getCurrentState()), -1)
                hyp = beam.getHyp(beam.getFinal())
                pred = beam.buildTargetTokens(hyp)[: self.beam_size]
                pred = [
                    torch.cat(
                        [x.view(-1) for x in p] + [zero] * (self.max_length - len(p))
                    ).view(1, -1)
                    for p in pred
                ]
                preds.append(torch.cat(pred, 0).unsqueeze(0))

            preds = torch.cat(preds, 0)
            return preds


class Beam(object):
    def __init__(self, size, sos, eos, device):
        self.size = size
        if device == "cuda":
            self.tt = torch.cuda
        elif device == "cpu":
            self.tt = torch
        # The score for each translation on the beam.
        self.scores = self.tt.FloatTensor(size).zero_()
        # The backpointers at each time-step.
        self.prevKs = []
        # The outputs at each time-step.
        self.nextYs = [self.tt.LongTensor(size).fill_(0)]
        self.nextYs[0][0] = sos
        # Has EOS topped the beam yet.
        self._eos = eos
        self.eosTop = False
        # Time and k pair for finished.
        self.finished = []

    def getCurrentState(self):
        "Get the outputs for the current timestep."
        batch = self.tt.LongTensor(self.nextYs[-1]).view(-1, 1)
        return batch

    def getCurrentOrigin(self):
        "Get the backpointers for the current timestep."
        return self.prevKs[-1]

    def advance(self, wordLk):
        """
        Given prob over words for every last beam `wordLk` and attention
        `attnOut`: Compute and update the beam search.

        Parameters:

        * `wordLk`- probs of advancing from the last step (K x words)
        * `attnOut`- attention at the last step

        Returns: True if beam search is complete.
        """
        numWords = wordLk.size(1)

        # Sum the previous scores.
        if len(self.prevKs) > 0:
            beamLk = wordLk + self.scores.unsqueeze(1).expand_as(wordLk)

            # Don't let EOS have children.
            for i in range(self.nextYs[-1].size(0)):
                if self.nextYs[-1][i] == self._eos:
                    beamLk[i] = -1e20
        else:
            beamLk = wordLk[0]
        flatBeamLk = beamLk.view(-1)
        bestScores, bestScoresId = flatBeamLk.topk(self.size, 0, True, True)

        self.scores = bestScores

        # bestScoresId is flattened beam x word array, so calculate which
        # word and beam each score came from
        prevK = bestScoresId // numWords
        self.prevKs.append(prevK)
        self.nextYs.append((bestScoresId - prevK * numWords))

        for i in range(self.nextYs[-1].size(0)):
            if self.nextYs[-1][i] == self._eos:
                s = self.scores[i]
                self.finished.append((s, len(self.nextYs) - 1, i))

        # End condition is when top-of-beam is EOS and no global score.
        if self.nextYs[-1][0] == self._eos:
            self.eosTop = True

    def done(self):
        return self.eosTop and len(self.finished) >= self.size

    def getFinal(self):
        if len(self.finished) == 0:
            self.finished.append((self.scores[0], len(self.nextYs) - 1, 0))
        self.finished.sort(key=lambda a: -a[0])
        if len(self.finished) != self.size:
            unfinished = []
            for i in range(self.nextYs[-1].size(0)):
                if self.nextYs[-1][i] != self._eos:
                    s = self.scores[i]
                    unfinished.append((s, len(self.nextYs) - 1, i))
            unfinished.sort(key=lambda a: -a[0])
            self.finished += unfinished[: self.size - len(self.finished)]
        return self.finished[: self.size]

    def getHyp(self, beam_res):
        """
        Walk back to construct the full hypothesis.
        """
        hyps = []
        for _, timestep, k in beam_res:
            hyp = []
            for j in range(len(self.prevKs[:timestep]) - 1, -1, -1):
                hyp.append(self.nextYs[j + 1][k])
                k = self.prevKs[j][k]
            hyps.append(hyp[::-1])
        return hyps

    def buildTargetTokens(self, preds):
        sentence = []
        for pred in preds:
            tokens = []
            for tok in pred:
                if tok == self._eos:
                    break
                tokens.append(tok)
            sentence.append(tokens)
        return sentence



from pathlib import Path
import configparser


class BadConfigError(Exception):
    pass


class DoclyConfig(object):

    def __init__(self, config_file):
        if Path(config_file).exists() and Path(config_file).is_file():
            self.config = configparser.ConfigParser()
            self.config.read(config_file)
            if "skipDirs" not in self.config:
                raise BadConfigError("\n\n=====>\nYou have mentioned a config file but it is badly formatted")
        else:
            self.config = None
    
    def _parent_contains_child(self, parent_path: str, child_path: str):
        child_path_rev = list(reversed(child_path.split("/")))
        parent_path_list = list(reversed(parent_path))
        parent_path_to_compare = parent_path_list[:len(child_path_rev)]
        return child_path_rev == parent_path_to_compare

    def is_dir_skipped(self, file_path_until_last_parent: str):
        if not self.config:
            return False
        for p in self.config["skipDirs"].keys():
            if self._parent_contains_child(file_path_until_last_parent, p):
                return True
        return False


class InputFeatures(object):
    """A single training/test features for a example."""
    def __init__(self,
                 example_id,
                 source_ids,
                 target_ids,
                 source_mask,
                 target_mask,

    ):
        self.example_id = example_id
        self.source_ids = source_ids
        self.target_ids = target_ids
        self.source_mask = source_mask
        self.target_mask = target_mask


def convert_examples_to_features(examples,
                                 tokenizer,
                                 max_source_length=256,
                                 max_target_length=128):
    features = []
    for example_index, example in enumerate(examples):
        source_tokens = tokenizer.tokenize(example.source)[:max_source_length-2]
        source_tokens = [tokenizer.cls_token] + source_tokens + [tokenizer.sep_token]
        source_ids =  tokenizer.convert_tokens_to_ids(source_tokens)
        source_mask = [1] * (len(source_tokens))
        padding_length = max_source_length - len(source_ids)
        source_ids += [tokenizer.pad_token_id] * padding_length
        source_mask += [0] * padding_length

        target_tokens = tokenizer.tokenize("None")
        target_tokens = [tokenizer.cls_token] + target_tokens + [tokenizer.sep_token]
        target_ids = tokenizer.convert_tokens_to_ids(target_tokens)
        target_mask = [1] * len(target_ids)
        padding_length = max_target_length - len(target_ids)
        target_ids += [tokenizer.pad_token_id] * padding_length
        target_mask += [0] * padding_length 

        features.append(
                InputFeatures(
                    example_index,
                    source_ids,
                    target_ids,
                    source_mask,
                    target_mask,
                )
            )
    return features



from rich import print as rprint


def print_on_console(text, color="green", emoji=None):
    if not emoji:
        rprint(f"[{color}]{text}[/{color}]")
    else:
        rprint(f"[{color}]{text}[/{color}]", f":{emoji}:")


from argparse import ArgumentParser
from pathlib import Path
import os
import shutil

from .args import setup_cmdline_args_for_docly_restore
from docly.ioutils.apply_diff import CACHE_DIR
from docly.ioutils import check_out_path, is_dir, is_python_file, query_yes_no, is_ipynb_notebook
from docly.ioutils.console_printer import print_on_console

parser = ArgumentParser()


def get_all_files_from_cache():
    return os.listdir(str(CACHE_DIR))


def main():
    setup_cmdline_args_for_docly_restore(parser)
    args = parser.parse_args()
    all_cached_files = get_all_files_from_cache()
    if not args.force:
        choice = query_yes_no("It will restore all MODIFIED files to the state of last run of `docly-gen`. Are you sure?")
    else:
        # Forceful application of restore command
        choice = True
    if choice:
        print_on_console("Restoring files", color="green")
        try:
            for file in check_out_path(Path().cwd().absolute()):
                if not is_dir(file) and (is_python_file(file) or is_ipynb_notebook(file)):
                    full_path = str(Path(file).absolute())
                    comparison_key = full_path[1:].replace("/", "#")
                    if comparison_key in all_cached_files:
                        source_file = str(CACHE_DIR / comparison_key)
                        final_file = str(Path(file).absolute())
                        shutil.move(source_file, final_file)
            print_on_console("Restoring done", color="green", emoji="thumbsup")
        except KeyboardInterrupt:
            print_on_console("Restoration not finished", color="red", emoji="X")


import random


class Example(object):
    """A single training/test example."""
    def __init__(self,
                 idx,
                 source,
                 target,
                 ):
        self.idx = idx
        self.source = source
        self.target = target


def make_example(code_tokens):
    idx = random.randint(1, 100000)
    examples = []
    for ct in code_tokens:
        ex = Example(idx=idx,
                    source=" ".join(ct),
                    target="")
        examples.append(ex)
    return examples


class NewExample(object):

    def __init__(self,
                 source,
                 target
                 ):
        self.source = source
        self.target = target


def make_new_example(code):
    return [NewExample(source=code, target=None)]



import os
import sys
from io import open

import numpy as np
import torch
import torch.nn as nn

from .example import make_example, make_new_example
from .input_features import convert_examples_to_features
from torch.utils.data import DataLoader, Dataset, SequentialSampler, TensorDataset

from transformers import (WEIGHTS_NAME, AdamW, get_linear_schedule_with_warmup,
                          RobertaConfig, RobertaModel, RobertaTokenizer)


MODEL_CLASSES = {'roberta': (RobertaConfig, RobertaModel, RobertaTokenizer)}

model_name_or_path = "microsoft/codebert-base"
beam_size = 10
max_target_length = 128
max_source_length = 256
seed = 42


def load_model(model_path, is_old=False):
    if is_old:
        from .model import Seq2Seq
    else:
        from .model_new import Seq2Seq
    
    config_class, model_class, tokenizer_class = MODEL_CLASSES['roberta']
    
    config = config_class.from_pretrained(model_name_or_path)
    if is_old:
        tokenizer = tokenizer_class.from_pretrained(model_name_or_path)
    else:
        tokenizer = tokenizer_class.from_pretrained(model_name_or_path, do_lower_case=False)

    encoder = model_class.from_pretrained(model_name_or_path, config=config)  
    decoder_layer = nn.TransformerDecoderLayer(d_model=config.hidden_size, 
                                               nhead=config.num_attention_heads)
    decoder = nn.TransformerDecoder(decoder_layer, num_layers=6)
    model = Seq2Seq(encoder=encoder,
                    decoder=decoder,
                    config=config,
                    beam_size=beam_size,
                    max_length=max_target_length,
                    sos_id=tokenizer.cls_token_id,
                    eos_id=tokenizer.sep_token_id
                   )
    if is_old:
        if not torch.cuda.is_available():
            model.load_state_dict(torch.load(model_path, map_location=torch.device('cpu')))
        else:
            model.load_state_dict(torch.load(model_path))
    else:
        if not torch.cuda.is_available():
            model.load_state_dict(torch.load(model_path, map_location=torch.device('cpu')), strict=False)
        else:
            model.load_state_dict(torch.load(model_path, strict=False))
    if not torch.cuda.is_available():
        model.to("cpu")
    model.eval()

    return model, tokenizer


def predict_docstring(model, tokenizer, code_tokens, is_old):
    examples = make_example(code_tokens) if is_old else make_new_example(code_tokens)

    features = convert_examples_to_features(examples, tokenizer)
    if is_old:
        all_source_ids = torch.tensor([f.source_ids for f in features], dtype=torch.long)
        all_source_mask = torch.tensor([f.source_mask for f in features], dtype=torch.long)
    else:
        all_source_ids = torch.tensor([f.source_ids[: max_source_length] for f in features], dtype=torch.long)
        all_source_mask = torch.tensor([f.source_mask[: max_source_length] for f in features], dtype=torch.long)

    eval_data = TensorDataset(all_source_ids, all_source_mask)

    eval_sampler = SequentialSampler(eval_data)
    batch_size = len(code_tokens) if is_old else len(eval_data)

    eval_dataloader = DataLoader(eval_data, sampler=eval_sampler, batch_size=batch_size)

    p=[]
    for batch in eval_dataloader:
        batch = tuple(t.to('cpu') for t in batch)
        source_ids, source_mask = batch

        with torch.no_grad():
            preds = model(source_ids=source_ids, source_mask=source_mask)

            for pred in preds:
                t=pred[0].cpu().numpy()
                t=list(t)
                if 0 in t:
                    t=t[:t.index(0)]
                text = tokenizer.decode(t,clean_up_tokenization_spaces=False)
                p.append(text)
    
    px = p[0].split()
    if px[-1] == ".":
        px[-2] = px[-2].strip() + "."
        px.pop()
    
    return [" ".join(px)]



from pathlib import Path

from docly.ioutils.console_printer import print_on_console


def main():
    print_on_console("Cleaning old model", color="green")
    if Path((Path.home() / ".docly" / "model" / "pytorch_model.bin")).exists():
        Path((Path.home() / ".docly" / "model" / "pytorch_model.bin")).unlink()
    print_on_console("Cleaning done", color="green", emoji="heavy_check_mark")



# Copyright (c) Microsoft Corporation. 
# Licensed under the MIT license.

import torch
import torch.nn as nn
import torch
from torch.autograd import Variable
import copy


class Seq2Seq(nn.Module):
    """
        Build Seqence-to-Sequence.
        
        Parameters:

        * `encoder`- encoder of seq2seq model. e.g. roberta
        * `decoder`- decoder of seq2seq model. e.g. transformer
        * `config`- configuration of encoder model. 
        * `beam_size`- beam size for beam search. 
        * `max_length`- max length of target for beam search. 
        * `sos_id`- start of symbol ids in target for beam search.
        * `eos_id`- end of symbol ids in target for beam search. 
    """
    def __init__(self, encoder,decoder,config,beam_size=None,max_length=None,sos_id=None,eos_id=None):
        super(Seq2Seq, self).__init__()
        self.encoder = encoder
        self.decoder=decoder
        self.config=config
        self.register_buffer("bias", torch.tril(torch.ones(2048, 2048)))
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)
        self.lsm = nn.LogSoftmax(dim=-1)
        self.tie_weights()
        
        self.beam_size=beam_size
        self.max_length=max_length
        self.sos_id=sos_id
        self.eos_id=eos_id
        
    def _tie_or_clone_weights(self, first_module, second_module):
        """ Tie or clone module weights depending of weither we are using TorchScript or not
        """
        if self.config.torchscript:
            first_module.weight = nn.Parameter(second_module.weight.clone())
        else:
            first_module.weight = second_module.weight
                  
    def tie_weights(self):
        """ Make sure we are sharing the input and output embeddings.
            Export to TorchScript can't handle parameter sharing so we are cloning them instead.
        """
        self._tie_or_clone_weights(self.lm_head,
                                   self.encoder.embeddings.word_embeddings)        
        
    def forward(self, source_ids=None,source_mask=None,target_ids=None,target_mask=None,args=None):   
        outputs = self.encoder(source_ids, attention_mask=source_mask)
        encoder_output = outputs[0].permute([1,0,2]).contiguous()
        if target_ids is not None:  
            attn_mask=-1e4 *(1-self.bias[:target_ids.shape[1],:target_ids.shape[1]])
            tgt_embeddings = self.encoder.embeddings(target_ids).permute([1,0,2]).contiguous()
            out = self.decoder(tgt_embeddings,encoder_output,tgt_mask=attn_mask,memory_key_padding_mask=(1-source_mask).bool())
            hidden_states = torch.tanh(self.dense(out)).permute([1,0,2]).contiguous()
            lm_logits = self.lm_head(hidden_states)
            # Shift so that tokens < n predict n
            active_loss = target_mask[..., 1:].ne(0).view(-1) == 1
            shift_logits = lm_logits[..., :-1, :].contiguous()
            shift_labels = target_ids[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = nn.CrossEntropyLoss(ignore_index=-1)
            loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1))[active_loss],
                            shift_labels.view(-1)[active_loss])

            outputs = loss,loss*active_loss.sum(),active_loss.sum()
            return outputs
        else:
            #Predict 
            preds=[]       
            zero=torch.LongTensor(1).fill_(0)     
            for i in range(source_ids.shape[0]):
                context=encoder_output[:,i:i+1]
                context_mask=source_mask[i:i+1,:]
                beam = Beam(self.beam_size,self.sos_id,self.eos_id)
                input_ids=beam.getCurrentState()
                context=context.repeat(1, self.beam_size,1)
                context_mask=context_mask.repeat(self.beam_size,1)
                for _ in range(self.max_length): 
                    if beam.done():
                        break
                    attn_mask=-1e4 *(1-self.bias[:input_ids.shape[1],:input_ids.shape[1]])
                    tgt_embeddings = self.encoder.embeddings(input_ids).permute([1,0,2]).contiguous()
                    out = self.decoder(tgt_embeddings,context,tgt_mask=attn_mask,memory_key_padding_mask=(1-context_mask).bool())
                    out = torch.tanh(self.dense(out))
                    hidden_states=out.permute([1,0,2]).contiguous()[:,-1,:]
                    out = self.lsm(self.lm_head(hidden_states)).data
                    beam.advance(out)
                    input_ids.data.copy_(input_ids.data.index_select(0, beam.getCurrentOrigin()))
                    input_ids=torch.cat((input_ids,beam.getCurrentState()),-1)
                hyp= beam.getHyp(beam.getFinal())
                pred=beam.buildTargetTokens(hyp)[:self.beam_size]
                pred=[torch.cat([x.view(-1) for x in p]+[zero]*(self.max_length-len(p))).view(1,-1) for p in pred]
                preds.append(torch.cat(pred,0).unsqueeze(0))
                
            preds=torch.cat(preds,0)                
            return preds   

        
class Beam(object):
    def __init__(self, size,sos,eos):
        self.size = size
        self.tt = torch
        # The score for each translation on the beam.
        self.scores = self.tt.FloatTensor(size).zero_()
        # The backpointers at each time-step.
        self.prevKs = []
        # The outputs at each time-step.
        self.nextYs = [self.tt.LongTensor(size)
                       .fill_(0)]
        self.nextYs[0][0] = sos
        # Has EOS topped the beam yet.
        self._eos = eos
        self.eosTop = False
        # Time and k pair for finished.
        self.finished = []

    def getCurrentState(self):
        "Get the outputs for the current timestep."
        batch = self.tt.LongTensor(self.nextYs[-1]).view(-1, 1)
        return batch

    def getCurrentOrigin(self):
        "Get the backpointers for the current timestep."
        return self.prevKs[-1]

    def advance(self, wordLk):
        """
        Given prob over words for every last beam `wordLk` and attention
        `attnOut`: Compute and update the beam search.

        Parameters:

        * `wordLk`- probs of advancing from the last step (K x words)
        * `attnOut`- attention at the last step

        Returns: True if beam search is complete.
        """
        numWords = wordLk.size(1)

        # Sum the previous scores.
        if len(self.prevKs) > 0:
            beamLk = wordLk + self.scores.unsqueeze(1).expand_as(wordLk)

            # Don't let EOS have children.
            for i in range(self.nextYs[-1].size(0)):
                if self.nextYs[-1][i] == self._eos:
                    beamLk[i] = -1e20
        else:
            beamLk = wordLk[0]
        flatBeamLk = beamLk.view(-1)
        bestScores, bestScoresId = flatBeamLk.topk(self.size, 0, True, True)

        self.scores = bestScores

        # bestScoresId is flattened beam x word array, so calculate which
        # word and beam each score came from
        prevK = bestScoresId // numWords
        self.prevKs.append(prevK)
        self.nextYs.append((bestScoresId - prevK * numWords))


        for i in range(self.nextYs[-1].size(0)):
            if self.nextYs[-1][i] == self._eos:
                s = self.scores[i]
                self.finished.append((s, len(self.nextYs) - 1, i))

        # End condition is when top-of-beam is EOS and no global score.
        if self.nextYs[-1][0] == self._eos:
            self.eosTop = True

    def done(self):
        return self.eosTop and len(self.finished) >=self.size

    def getFinal(self):
        if len(self.finished) == 0:
            self.finished.append((self.scores[0], len(self.nextYs) - 1, 0))
        self.finished.sort(key=lambda a: -a[0])
        if len(self.finished) != self.size:
            unfinished=[]
            for i in range(self.nextYs[-1].size(0)):
                if self.nextYs[-1][i] != self._eos:
                    s = self.scores[i]
                    unfinished.append((s, len(self.nextYs) - 1, i)) 
            unfinished.sort(key=lambda a: -a[0])
            self.finished+=unfinished[:self.size-len(self.finished)]
        return self.finished[:self.size]

    def getHyp(self, beam_res):
        """
        Walk back to construct the full hypothesis.
        """
        hyps=[]
        for _,timestep, k in beam_res:
            hyp = []
            for j in range(len(self.prevKs[:timestep]) - 1, -1, -1):
                hyp.append(self.nextYs[j+1][k])
                k = self.prevKs[j][k]
            hyps.append(hyp[::-1])
        return hyps
    
    def buildTargetTokens(self, preds):
        sentence=[]
        for pred in preds:
            tokens = []
            for tok in pred:
                if tok==self._eos:
                    break
                tokens.append(tok)
            sentence.append(tokens)
        return sentence



from pathlib import Path

from tree_hugger.core import PythonParser


def set_up_parser():
    return 


def parse(file_path: str, tslib_path):
    python_parser = PythonParser(library_loc=tslib_path)
    return python_parser.parse_file(file_path), python_parser


def get_func_body_and_docstr(python_parser: PythonParser):
    res = python_parser.get_all_function_bodies(strip_docstr=True, get_index=True)
    for key, value in res.items():
        yield key, value


import os
from pathlib import Path
import platform

from docly.ioutils import download_from_url
from docly.ioutils.console_printer import print_on_console

SUPPORTED_PLATFORMS = ["linux", "darwin"]


def inspect_and_download_latest_model(model_root: Path, download_url: str, is_old=False) -> bool:
    model_file = "pytorch_model_new.bin" if not is_old else "pytorch_model.bin"

    if not model_root.exists():
        os.makedirs(str(model_root))
        os.mkdir(str(model_root / "model"))
    elif model_root.exists() and not (model_root / "model").exists():
        os.mkdir(str(model_root / "model"))

    if Path(model_root/ "model" / model_file).exists() and Path(model_root/ "model" / model_file).is_file():
        return True
    
    print_on_console("There is no model. Downloading (maybe because you chose to use the newer model)", color="green")
    download_from_url(download_url, str(Path(model_root/ "model" / model_file)))
    print_on_console("Download complete", color="green", emoji="heavy_check_mark")
    return True


def inspect_and_download_latest_tslibs(tslibs_root: Path, download_url: str) -> bool:
    os_name = platform.system().lower()
    if os_name not in SUPPORTED_PLATFORMS:
        return (False, None)
    
    file_name = "python_ts_darwin64.so" if os_name == "darwin" else "python_ts_nix64.so"
    download_url = f"{download_url}{file_name}"

    if (tslibs_root/ "tslibs" / file_name).exists() and (tslibs_root/ "tslibs" / file_name).is_file():
        return (True, file_name) 

    if not (tslibs_root / "tslibs").exists():
        os.mkdir(str(tslibs_root / "tslibs"))
    
    print_on_console("There is no tree-sitter lib. Downloading", color="green")
    download_from_url(download_url, str(Path(tslibs_root/ "tslibs" / file_name)))
    print_on_console("Download complete", color="green", emoji="heavy_check_mark")
    return (True, file_name)
    


import pathlib
from setuptools import find_packages, setup

from docly import __version__

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


setup(
name='docly',
description="Generate docstrings for your python functions. Automatically!",
long_description=README,
long_description_content_type="text/markdown",
url="https://github.com/autosoft-dev/docly",
author="CodistAI",
author_email="shubhadeep@codist-ai.com",
include_package_data=True,
license="MIT",
classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
],
version=__version__,
packages=find_packages(exclude=("tests",)),
install_requires=["tqdm", "requests", "torch", 
                  "pyfiglet", "rich", "dpu-utils", "numpy", 
                  "nltk", "transformers==2.5.0", "tree_hugger",
                  "halo", "invoke",
                  ],
extras_require={
        "jupyter": ["jupytext"]
},
entry_points = {
    'console_scripts': ['docly-gen=docly.cli.docly_gen:main',
                        'docly-restore=docly.cli.docly_restore:main',
                        'docly-clean=docly.cli.docly_clean:main'],
},
)



import os
from pathlib import Path
import requests
import shutil
import sys
from distutils.version import LooseVersion
import time

from tqdm import tqdm

from docly.parser import parser as py_parser
from docly.tokenizers import tokenize_code_string
from docly import __version__

# from c2nl.objects import Code
UPDATE_CHECK_URL = "http://3.80.2.138:8584/vercheck/check-version/"
# UPDATE_CHECK_URL = "http://127.0.0.1:5000/vercheck/check-version/"

interaction_cache = lambda : Path(Path.home() / ".docly" / "interaction_cache")

CACHE_DIR = (Path().home() / ".docly" / "file_cache")

cache_exists = lambda : CACHE_DIR.exists()
make_cache_dir = lambda : os.mkdir(str(CACHE_DIR))


def _compare_installed_version_with_latest(v1, v2):
    try:
        current_version = LooseVersion(v1)
        latest_version = LooseVersion(v2)
        assert current_version == latest_version
        return True
    except AssertionError:
        return False


def look_for_update():
    with requests.sessions.Session() as s:
        try:
            r = s.get(UPDATE_CHECK_URL, timeout=2)
            r.raise_for_status()
            if not _compare_installed_version_with_latest(__version__, r.text):
                i_c = interaction_cache()
                return True
            return False
        except Exception:
            i_c = interaction_cache()
            if not i_c.exists():
                os.mkdir(i_c)
            if not (i_c / "icache.txt").exists():
                with open((i_c / "icache.txt"), "w") as f:
                    f.write(str(int(time.time())) + "\n")
            else:
                with open((i_c / "icache.txt"), "a") as f:
                    f.write(str(int(time.time())) + "\n")
            return False


def is_dir(base_path):
    if isinstance(base_path, Path):
        return base_path.is_dir()
    elif isinstance(base_path, str):
        return Path(base_path).is_dir()
    else:
        return False


def is_python_file(file_path):
    if isinstance(file_path, Path):
        return file_path.suffix == ".py"
    elif isinstance(file_path, str):
        return Path(file_path).suffix == ".py"
    else:
        return False


def is_ipynb_notebook(file_path):
    if isinstance(file_path, Path):
        return file_path.suffix == ".ipynb"
    elif isinstance(file_path, str):
        return Path(file_path).suffix == ".ipynb"
    else:
        return False


def download_from_url(url, dst):
    """
    @param: url to download file
    @param: dst place to put the file
    """
    file_size = int(requests.head(url).headers["Content-Length"])
    if os.path.exists(dst):
        first_byte = os.path.getsize(dst)
    else:
        first_byte = 0
    if first_byte >= file_size:
        return file_size
    header = {"Range": "bytes=%s-%s" % (first_byte, file_size)}
    pbar = tqdm(
        total=file_size, initial=first_byte,
        unit='B', unit_scale=True, desc=dst.split('/')[-1])
    req = requests.get(url, headers=header, stream=True)
    with(open(dst, 'ab')) as f:
        for chunk in req.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
                pbar.update(1024)
    pbar.close()
    return file_size


def check_out_path(target_path: Path):
    """"
    This function recursively yields all contents of a pathlib.Path object
    """
    yield target_path
    for file in target_path.iterdir():
        if file.is_dir():
            yield from check_out_path(file)
        else:
            yield file.absolute()
    

def process_file(file_path: Path, ts_lib_path: str, use_old=False):
    result, parser_obj = py_parser.parse(file_path, ts_lib_path)
    func_and_params = parser_obj.get_all_function_names_with_params()
    if result:
        for func_name, data in py_parser.get_func_body_and_docstr(parser_obj):
            # print(py_toeknizer.tokenize_code_string(func_body))
            # code.tokens = tokenizer.tokenize(func_body).data
            # code.text = func_body
            (func_body, docstr), start, end = data
            ret_start = (start[0]+1, start[1])
            params = func_and_params[func_name]

            code_str = [tokenize_code_string(func_body)] if use_old else func_body

            yield code_str, params, ret_start, func_name, docstr.strip()


def query_yes_no(question, default="yes"):
    """Ask a yes/no question and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes", "no", or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '{}}'".format(default))

    while True:
        print(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


from tokenize import tokenize
from io import BytesIO


def tokenize_code_string(text):
    code_tokens = []
    for tok in tokenize(BytesIO(text.encode('utf-8')).readline):
        if tok.string.strip() != "" and tok.string.strip() != "utf-8":
            code_tokens.append(tok.string.strip().lower())
    return code_tokens


from pathlib import Path
import shutil
import logging

from invoke import run
from docly.ioutils import CACHE_DIR, cache_exists, make_cache_dir

try:
    import jupytext
    JUPYTEXT_AVAILABLE = True
except Exception:
    JUPYTEXT_AVAILABLE = False


IPYNB_TO_PY_CMD = "jupytext --set-formats ipynb,py:percent"
PY_TO_IPYNB_CMD = "jupytext --update --to ipynb"


def convert_ipynb_to_python(notebook_path: Path):
    if JUPYTEXT_AVAILABLE:
        try:
            jupytext.read(notebook_path.absolute())
        except Exception:
            return None

    actual_file_path = str(notebook_path.absolute())
    cache_file_name = actual_file_path[1:].replace("/", "#")

    if not cache_exists():
        make_cache_dir()
    
    if (CACHE_DIR / cache_file_name).exists():
        (CACHE_DIR / cache_file_name).unlink()
    
    shutil.copy(actual_file_path, str(CACHE_DIR / cache_file_name))

    result = run(f"{IPYNB_TO_PY_CMD} {str(notebook_path.absolute())}", hide=True, warn=True)
    
    if not result.ok:
        logging.error("Could not run the conversion command. Are you using an old version of Jupyter notebook? Otherwise, Maybe use `pip install 'docly[jupyter]'")
        return None
    else:
        return (notebook_path.absolute().parent / (notebook_path.stem + '.py')).absolute()


def convert_python_to_ipynb(python_file_path: Path):
    result = run(f"{PY_TO_IPYNB_CMD} {str(python_file_path.absolute())}", hide=True, warn=True)
    if not result.ok:
        logging.error("Could not run the conversion command. Maybe use `pip install 'docly[jupyter]'")
    else:
        return (python_file_path.absolute().parent / (python_file_path.stem + '.ipynb')).absolute()


from argparse import ArgumentParser
from argparse import ArgumentDefaultsHelpFormatter
from pathlib import Path


def setup_cmdline_args_for_docly_gen(parser: ArgumentParser):
    parser.add_argument("--no_generate_diff", action="store_false", 
                        help="Do not generate the diff. Only prints a report on the console")
    parser.add_argument("--no_generate_args_list", action="store_false",
                        help="Do not generate argument list in the docstring")
    parser.add_argument("--no_print_report", action="store_false",
                        help="Do not prompt to show the report once the diff is generated")
    parser.add_argument("--run_on_notebooks", action="store_true",
                        help="If you want docly to run on notebook (.ipynb) files (Requires jupytext and defaults false)")
    parser.add_argument("--docstring_style", type=str, default="google", 
                        help="What style of docstring you want [google, numpy, sphinx]. Defaults to `google` style.")
    parser.add_argument("--config_file", type=str, default="docly-config.ini",
                        help="Configuration file for docly")
    parser.add_argument("--use-old-model", action="store_true",
                        help="Do you want to run the older model? Saves a second download, but bad quality")
    parser.add_argument("--force", action="store_true",
                        help="Applies the changes without an interactive propmt")
    # This is manmdatory
    parser.add_argument("files", type=str, nargs="+",
                         help="List the files/dirs you want to run it on")


def setup_cmdline_args_for_docly_restore(parser: ArgumentParser):
    parser.add_argument("--force", action="store_true", 
                        help="Disables interactive restoring")